#import io
import os
from glob import glob

import yaml

from .backupserver import BackupServer
from .utils import dd, dump


class Backups:
    """Backup a complete set of servers
    """
    __author__ = "Matthew Reschke <mail@mreschke.com>"
    __license__ = "MIT"

    def __init__(self, *, servers=None, defaults=None, config_path='/etc/serverbackups'):
        # NOTICE: Do NOT log anything in this class.  Too early or it will log even during --help

        # Blank attributes
        self.config_path = None
        self.configd_path = None
        self.defaults_file = None

        # Get defaults from parameter, or config path defaults.yml
        self.defaults = self.__get_defaults(servers, defaults, config_path)

        # Get servers from parameter or config path .d folder
        servers = self.__get_servers(servers)

        # Build complete servers dictionary by merging each server with defaults
        self.servers = {}
        for server in servers:
            self.servers[server] = self.__merge_defaults(servers[server], self.defaults)

    def run(self, server=None, cluster=None):
        if server:
            # Backup a single server from the config
            if server in self.servers:
                BackupServer(server, self.servers[server], self.defaults).run()
        elif cluster:
            # Backup all servers marked with this cluster
            for server, options in self.servers.items():
                if options['cluster'] == cluster:
                    BackupServer(server, options, self.defaults).run()
        else:
            for server, options in self.servers.items():
                BackupServer(server, options, self.defaults).run()

    def __get_defaults(self, servers, defaults, config_path):
        """Get defaults from parameter, or config path defaults.yml
        """
        if servers and defaults:
            # Using passed in defaults
            return defaults
        elif servers and not defaults:
            # No defaults, so everything is self-contained in servers
            # But still need to create a blank defaults so our merge works properly

            # Using serverbackup builtint default.yml
            file = os.path.dirname(os.path.realpath(__file__)) + '/templates/' + 'defaults_blank.yml'
            default_config = ''
            with open(file, 'r') as stream:
                default_config = yaml.safe_load(stream)
            return default_config
        else:
            # No servers defined, using config path
            self.config_path = config_path
            self.configd_path = config_path + '/config.d'
            self.defaults_file = config_path + '/defaults.yml'

            # Read defaults.yml file
            if os.path.exists(self.defaults_file):
                defaults = {}
                with open(self.defaults_file, 'r') as stream:
                    defaults = yaml.safe_load(stream)
                return defaults
            else:
                print("Defaults file " + self.defaults_file + " does not exist")
                exit(1)

    def __get_servers(self, servers):
        """Get servers from parameter or config path .d folder
        """
        if not servers:
            # Read each servers .yml config in config.d folder
            servers = {}
            files = glob(self.configd_path + "/*.yml")
            for file in files:
                with open(file, 'r') as stream:
                    servers = {**servers, **yaml.safe_load(stream)}
        return servers

    def __merge_defaults(self, options, defaults):
        """Properly merge or replace options with defaults"""

        # Ensure options has all keys
        options.setdefault('cluster', '')
        options.setdefault('prune', {})
        options.setdefault('rsync', {})
        options.setdefault('source', {})
        options.setdefault('destination', {})
        options['source'].setdefault('ssh', {})
        options['destination'].setdefault('ssh', {})
        options.setdefault('backup', {})
        options['backup'].setdefault('preScripts', {})
        options['backup'].setdefault('files', {})
        options['backup'].setdefault('mysql', {})
        options['backup'].setdefault('postScripts', {})
        options['backup']['files'].setdefault('common', [])
        options['backup']['files'].setdefault('extra', [])
        options['backup']['files'].setdefault('exclude', [])
        options['backup']['mysql'].setdefault('excludeDbs', [])

        # Merge arrays
        options['backup']['files']['common'] = defaults['backup']['files']['common'] + options['backup']['files']['common']
        options['backup']['files']['extra'] = defaults['backup']['files']['extra'] + options['backup']['files']['extra']
        options['backup']['files']['exclude'] = defaults['backup']['files']['exclude'] + options['backup']['files']['exclude']
        options['backup']['mysql']['excludeDbs'] = defaults['backup']['mysql']['excludeDbs'] + options['backup']['mysql']['excludeDbs']

        # Replace dictionaries
        options['prune'] = {**defaults['prune'], **options['prune']}
        options['rsync'] = {**defaults['rsync'], **options['rsync']}
        options['source']['ssh'] = {**defaults['source']['ssh'], **options['source']['ssh']}
        options['source'] = {**defaults['source'], **options['source']}
        options['destination']['ssh'] = {**defaults['destination']['ssh'], **options['destination']['ssh']}
        options['destination'] = {**defaults['destination'], **options['destination']}

        options['backup']['mysql'] = {**defaults['backup']['mysql'], **options['backup']['mysql']}

        options['backup']['preScripts'] = {**defaults['backup']['preScripts'], **options['backup']['preScripts']}
        for script in options['backup']['preScripts'].keys():
            if script in defaults['backup']['preScripts'].keys():
                options['backup']['preScripts'][script] = {**defaults['backup']['preScripts'][script], **options['backup']['preScripts'][script]}

        options['backup']['postScripts'] = {**defaults['backup']['postScripts'], **options['backup']['postScripts']}
        for script in options['backup']['postScripts'].keys():
            if script in defaults['backup']['postScripts'].keys():
                options['backup']['postScripts'][script] = {**defaults['backup']['postScripts'][script], **options['backup']['postScripts'][script]}

        # Finally replace remaining flat values
        options = {**defaults, **options}
        return options



import os
import shutil
import subprocess
from datetime import date, datetime, timedelta
from types import SimpleNamespace as obj

from . import log
from .utils import dd, dump


class BackupServer:
    """Backup a single server defined as a dictionary of options"""

    __author__ = "Matthew Reschke <mail@mreschke.com>"
    __license__ = "MIT"

    def __init__(self, server, options, defaults=None):
        """Initialize backup class
        """
        # This class assumes a perfact and complete options dictionary for ONE single server
        # All defaults are already merged by the handling Backups class before being passed in
        self.options = options

        # Make option attributes for easier usage
        self.enabled = options['enabled']
        self.cluster = options['cluster']
        self.prune = obj(**options['prune'])
        self.rsync = obj(**options['rsync'])
        self.rsync.exclude_file = '/tmp/backups.exclude'
        self.src = obj(**options['source'])
        self.src.ssh = obj(**options['source']['ssh'])
        self.dest = obj(**options['destination'])
        self.dest.ssh = obj(**options['destination']['ssh'])
        self.files = obj(**options['backup']['files'])
        self.pre_scripts = []
        if 'preScripts' in options['backup']:
            for script in options['backup']['preScripts']:
                self.pre_scripts.append(obj(**{
                    'name': script,
                    **options['backup']['preScripts'][script]
                }))
        self.post_scripts = []
        if 'postScripts' in options['backup']:
            for script in options['backup']['postScripts']:
                self.post_scripts.append(obj(**{
                    'name': script,
                    **options['backup']['postScripts'][script]
                }))
        self.mysql = obj(**options['backup']['mysql'])

        # Set other attributes
        self.src.server = server
        self.dest.server = self.src.server if self.dest.location == 'local' else self.dest.ssh.host
        #self.dest.snapshot = datetime.today().strftime(options['snapshotDateFormat'])
        self.dest.snapshot = datetime.today().strftime("%Y-%m-%d_%H%M%S")
        self.dest.snapshot_path = self.path('snapshots/' + self.dest.snapshot)
        self.today = datetime.today().strftime("%Y%m%d")

    def run(self):
        """Run backups
        """
        # Do not run if backup is disabled
        if not self.enabled:
            log.notice("Server {} configuration DISABLED, skipping server".format(self.src.server))
            return

        # Begin backup routine
        log.line()
        log.header2("Backing up server {}".format(self.src.server))
        log.line()
        log.bullet2("Direction {} -> {}".format(self.src.location, self.dest.location))
        log.bullet2("SRC server={}, type={}".format(self.src.server, self.src.location))
        log.bullet2("DEST server={}, type={}".format(self.dest.server, self.dest.location))

        # Show YAML config in output
        log.bullet2("YAML config for {}".format(self.src.server))
        log.info((self.options))
        dump(self.options)  # Keep this one, nice output

        # Prepare backup directories
        self.prepare()

        # Run pre-scripts
        self.backup_scripts('pre', self.pre_scripts)

        # Backup files
        self.backup_files()

        # Backup MySQL
        self.backup_mysql()

        # Run post-scripts
        self.backup_scripts('post', self.post_scripts)

        # Cleanup old snapshots
        self.cleanup()

    def prepare(self):
        """Prepare backup system
        """
        log.header4("Preparing System")
        # Create DEST snapshot directory
        log.bullet("Creating DEST snapshot folder {}:{}".format(self.dest.server, self.dest.snapshot_path))
        self.execute_dest("mkdir -p " + self.dest.snapshot_path)

    def backup_scripts(self, type, scripts):
        """Execute script on DEST and optionally backup the output to snapshot folder
        """
        for script in scripts:
            if not script.enabled: continue
            log.header4("Running {} script '{}'".format(type.upper(), script.name))
            output = None
            if hasattr(script, 'output'):
                log.bullet("Saving script output to '{}' to DEST snapshot".format(script.output))
                output = self.dest.snapshot_path + '/' + script.output
            self.execute(cmd=script.script, outfile=output)

    def backup_files(self):
        """Backup local or remote files using rsync hardlink snapshots
        See http://anouar.adlani.com/2011/12/how-to-backup-with-rsync-tar-gpg-on-osx.html
        """
        log.header4("Backing up files")

        # Merge all files into a single array
        files = self.options['backup']['files']['common'] + self.options['backup']['files']['extra']
        if (not files): return

        # Create exclude file for rsync
        if os.path.exists(self.rsync.exclude_file): os.remove(self.rsync.exclude_file)
        with open(self.rsync.exclude_file, 'w') as f: f.write('\n'.join(self.options['backup']['files']['exclude']))

        # Get a string of source files.  This means many files can be backed up
        # with a single rsync command
        src = []
        for file in files:
            if self.src.location == 'local':
                src.append(file)
            elif self.src.location == 'ssh':
                src.append(self.src.ssh.user + "@" + self.src.ssh.host + ":" + file)
        src = ' '.join(src)  # Array to string rsync

        # Get destination string
        dest = self.path()
        if self.dest.location == 'ssh':
            dest = self.dest.ssh.user + "@" + self.dest.ssh.host + ":" + dest

        # Snapshot folder (from dest, so either local or user@server remote)
        snapshot = dest + '/snapshots/' + self.dest.snapshot  # Will be either local or user@server

        # Current folder for rsync --link-dest, even if rsyncing to remove server this is a / local style path
        # not user@server.  Rsync is smart enough to know you mean the remote servers filesystem here.
        current = self.path('current')
        #dd(src, dest, current, snapshot)

        # Backup new snapshot
        log.bullet("Backing up files {} to {}".format(src, snapshot))
        params = "--archive --relative --hard-links --delete --progress --partial --compress "
        if self.rsync.verbose:
            params += "--verbose "
        else:
            params += "--quiet "

        ssh_string = ''
        if self.src.location == 'ssh':
            ssh_string = "-e \"ssh -p {} -i {} -o LogLevel=quiet\"".format(self.src.ssh.port, self.src.ssh.key)
        elif self.dest.location == 'ssh':
            ssh_string = "-e \"ssh -p {} -i {} -o LogLevel=quiet\"".format(self.dest.ssh.port, self.dest.ssh.key)

        # Execute rsync
        try:
            cmd = "rsync {} {} --exclude-from={} --link-dest={} {} {}".format(ssh_string, params, self.rsync.exclude_file, current, src, snapshot)

            # Always executed local (where backup script is run)
            log.bullet4(cmd)
            os.system(cmd)

        except KeyboardInterrupt:
            exit()

    def backup_mysql(self):
        """Backup mysql databases and tables
        """
        log.header4("Backing up MySQL databases")

        # Ignore disabled mysql entries
        if not self.mysql.enabled: return

        # Always local, the self.execute below will add proper SSH output as needed
        dest = self.dest.snapshot_path + '/mysqldump'
        self.execute_dest("mkdir -p " + dest)

        # Mysql Connection
        mysql_cmd = self.mysql.mysqlCmd
        dump_cmd = self.mysql.dumpCmd
        host = self.mysql.host
        port = self.mysql.port
        user = self.mysql.user
        password = self.mysql.password
        dbs = self.mysql.dbs
        excludeDbs = self.mysql.excludeDbs

        # Convert dbs string or array into proper array of {name, tables[]} dictionary
        if isinstance(dbs, str):
            if (dbs == '*'):
                # Convert * into dictionary of all tables on system
                cmd = mysql_cmd + " --user={} --password='{}' --host={} --port={} -Bse 'show databases'".format(user, password, host, port)
                databases = self.execute(cmd=cmd, output_list=True, skip_logging=True)  # Skip logging as it has password
                databases = set(databases) - set(excludeDbs)  # Also dedups, now a set not a list
                dbs = []
                for database in databases:
                    dbs.append({'name': database, 'tables': '*'})
            else:
                dbs = [{'name': dbs, 'tables': '*'}]
        elif isinstance(dbs, list) and isinstance(dbs[0], dict):
            # Do nothing, dbs is in exact dictionary format we need
            # Just end the if
            pass
        elif isinstance(dbs, list):
            tmpDbs = []
            for db in dbs:
                tmpDbs.append({'name': db, 'tables': '*'})
            dbs = tmpDbs

        # Backup databases and tables
        for db in dbs:
            name = db['name']
            tables = db['tables']
            if tables == '*':
                log.bullet("Backing up MySQL database '{}', ALL tables".format(name))
                tables = ''
            else:
                log.bullet("Backing up MySQL database '{}', ONLY {} tables".format(name, str(len(tables))))
                tables = ' '.join(tables)

            # Backup database and all or some tables
            # --quick takes Retrieve rows for a table from the server a row at a time instead of reading entire table in memory, good for large tables
            # --single-transaction Issue a BEGIN SQL statement before dumping data from server.  Uses a consistent read and guarantees that data seen by muysqldump does nto change
            # --flush-logs Flush MySQL server log files before starting dump
            # --master-data Write the binary log file name and position to the output, good if this is part of a cluster

            # Dump flags are optionaly even in defaults.  I only added because MariaDB 10.1 does NOT like flush-logs
            flags = "--quick --single-transaction --flush-logs"
            if hasattr(self.mysql, 'dumpFlags'):
                flags = self.mysql.dumpFlags

            # Output is piped ON MySQL server to gzip before being sent over SSH!  Do not use the --compress option, that is only between client and server (which is localhost anyhow usually)
            cmd = dump_cmd + " --user={} --password='{}' --host={} --port={} {} {} {} | gzip".format(user, password, host, port, flags, name, tables)
            self.execute(cmd=cmd, outfile=dest + '/' + name + ".sql.gz", skip_logging=True, dryrun=False)

    def cleanup(self):
        """Cleanup old snapshots
        """
        log.header4("Cleaning up system and symlinking current snapshot")

        # Delete rsync exclude file
        if os.path.exists(self.rsync.exclude_file): os.remove(self.rsync.exclude_file)

        # Define backup directories
        current = self.path('current')
        snapshots = self.path('snapshots')

        # Symlink latest snapshot to current link
        # Only difference is on remote, the $() is escaped
        if self.dest.location == 'local':
            cmd = "ln -snf $(ls -1d {}/* | tail -n1) {}".format(snapshots, current)
        else:
            cmd = "ln -snf \\$(ls -1d {}/* | tail -n1) {}".format(snapshots, current)
        self.execute_dest(cmd)

        # Prune old snapshots according to daily, weekly, monthly and yearly prune variables
        self.prune_snapshots()

    def prune_snapshots(self):
        """Prune old snapshots according to daily, weekly, monthly and yearly prune variables
        """
        log.header4("Pruning Snapshots (keepDaily:{}, keepWeekly:{}, keepMonthly:{}, keepYearly:{})".format(self.prune.keepDaily, self.prune.keepWeekly, self.prune.keepMonthly, self.prune.keepYearly))

        # Define snapshot date parser helper
        def dates(snapshot):
            # Snapshot date and snapshot time, and weekday
            sd = date(int(snapshot[0:4]), int(snapshot[5:7]), int(snapshot[8:10]))
            st = int(snapshot[11:-1])
            wd = int(sd.strftime("%w"))  # 0=Sunday, 6=Saturday
            return (sd, st, wd)

        # Snapshot folder and start date
        snapshot_path = self.path('snapshots')
        today = date.today()

        # FAKER DATA for Testing Only
        # start = date(2009, 1, 1)
        # while start < today:
        #     start += timedelta(days=1)
        #     snapshot = snapshot_path + '/' + start.strftime("%Y-%m-%d_010101")
        #     if not os.path.exists(snapshot): os.mkdir(snapshot)
        #     snapshot = snapshot_path + '/' + start.strftime("%Y-%m-%d_020101")
        #     if not os.path.exists(snapshot): os.mkdir(snapshot)

        # Get all snapshot folders (always on destination)
        snapshots = sorted(self.execute_dest("/bin/ls " + snapshot_path))
        if len(snapshots) == 0: return

        # Find days to keep
        # Keeps multiple of the same day back to keepDaily
        keep_daily = []
        for i, snapshot in enumerate(snapshots):
            # Get snapshot date and time
            sd, st, wd = dates(snapshot)

            days_back = today - timedelta(days=self.prune.keepDaily)
            if sd >= days_back:
                keep_daily.append(snapshot)

        # Remove keep_daily from snapshots
        snapshots = [x for x in snapshots if x not in keep_daily]
        if len(snapshots) == 0: return

        # Get list of only latest snapshot of each day
        # Remove any snapshots of the SAME day that are earlier in the day, keeping the latest of each day
        add = ''
        current = dates(snapshots[0])[0]  # date
        latest_days = []
        for i, snapshot in enumerate(snapshots):
            # Get snapshot date and time
            sd, st, wd = dates(snapshot)

            if sd != current:
                latest_days.append(add)
            add = snapshot
            current = sd
        latest_days.append(snapshots[-1])
        snapshots = latest_days

        # Get list of only latest snapshots of each week
        add = ''
        current = dates(snapshots[0])[2]  # week day
        latest_weeks = []
        for i, snapshot in enumerate(snapshots):
            # Get snapshot date and time
            sd, st, wd = dates(snapshot)

            if wd < current:
                latest_weeks.append(add)
            add = snapshot
            current = wd

        # Take top latest weekly snapshots according to self.prune.keepWeekly
        keep_weekly = latest_weeks[-self.prune.keepWeekly:]

        # Remove keep_weekly from snapshots
        snapshots = [x for x in snapshots if x not in keep_weekly]
        if len(snapshots) == 0: return

        # Get list of only latest snapshots of each month
        add = ''
        current = dates(snapshots[0])[0].day
        latest_months = []
        for i, snapshot in enumerate(snapshots):
            # Get snapshot date and time
            sd, st, wd = dates(snapshot)

            if sd.day < current:
                latest_months.append(add)
            add = snapshot
            current = sd.day

        # Take top latest monthly snapshots according to self.prune.keepMonthly
        keep_monthly = latest_months[-self.prune.keepMonthly:]

        # Remove keep_monthly from snapshots
        snapshots = [x for x in snapshots if x not in keep_monthly]
        if len(snapshots) == 0: return

        # Get list of only latest snapshots of each year
        add = ''
        current = dates(snapshots[0])[0].month
        latest_years = []
        for i, snapshot in enumerate(snapshots):
            # Get snapshot date and time
            sd, st, wd = dates(snapshot)

            if sd.month < current:
                latest_years.append(add)
            add = snapshot
            current = sd.month

        # Take top latest yearly snapshots according to self.prune.keepYearly
        keep_yearly = latest_years[-self.prune.keepYearly:]

        # Remove keep_monthly from snapshots
        snapshots = [x for x in snapshots if x not in keep_yearly]
        if len(snapshots) == 0: return

        # Now we have a perfect set of snapshots to keep
        # dump('KEEP DAILY', keep_daily)
        # dump('KEEP WEEKLY', keep_weekly)
        # dump('KEEP MONTHLY', keep_monthly)
        # dump('KEEP YEARLY', keep_yearly)

        # Delete all other snapshot folders besides those in keep_*
        # Have to get a fresh list of actual files and subtract daily, weekly, monthly, yearly
        # Cannot simply use snapshots as is not, because we removed all files of the same day
        oldsnapshots = sorted(self.execute_dest("/bin/ls " + snapshot_path))
        oldsnapshots = [x for x in oldsnapshots if x not in keep_daily]
        oldsnapshots = [x for x in oldsnapshots if x not in keep_weekly]
        oldsnapshots = [x for x in oldsnapshots if x not in keep_monthly]
        oldsnapshots = [x for x in oldsnapshots if x not in keep_yearly]
        for oldsnapshot in oldsnapshots:
            fullpath = snapshot_path + '/' + oldsnapshot
            log.bullet("Deleting old snapshot {}".format(fullpath))
            self.execute_dest("/bin/rm -rf '{}'".format(fullpath))

    def execute(self, cmd, outfile=None, output_list=False, skip_logging=False, dryrun=False):
        """Execute command local or remote with optional output saved to snapshot folder
        local->local = script=local, output=local
        local->remote = script=local, output=remote
        remote->local = script=remote, output=local
        """
        # Get src and dest location types (local or ssh)
        src = self.src.location
        dest = self.dest.location
        outfile_cmd = ''

        # local->local = script=local, output=local
        if src == 'local' and dest == 'local':
            # Execute script locally, save output locally
            # cat /etc/hosts > /snapshot/dir/hosts
            if outfile: outfile_cmd = cmd + " > " + str(outfile)
        elif src == 'local' and dest == 'ssh':
            # Execute script locally, save output remotely
            # cat /etc/hosts | ssh toor@linstore 'cat > /snapshot/dir/hosts'
            if outfile: outfile_cmd = cmd + " | " + self.ssh_string('dest') + "'cat > " + str(outfile) + "'"
        elif src == 'ssh' and dest == 'local':
            # Execute script remotely, save output locally
            #ssh toor@linstore cat /etc/hosts > /snapshot/dir/hosts
            cmd = self.ssh_string('src') + " \"" + cmd + "\""
            if outfile: outfile_cmd = cmd + " >  " + str(outfile)

        if dryrun:
            cmd = 'DRYRUN: ' + cmd
            outfile_cmd = 'DRYRUN: ' + outfile_cmd

        if outfile is None and output_list is False:
            # Execute the command, no output
            if not skip_logging: log.bullet4(cmd)
            if not dryrun: os.system(cmd)
        else:
            if outfile:
                # Execute the command and > output to outfile
                if not skip_logging: log.bullet4(outfile_cmd)
                if not dryrun: os.system(outfile_cmd)
            elif output_list:
                # Execute the command and capture output to python list
                if not skip_logging: log.bullet4(cmd)
                if not dryrun: return subprocess.run(cmd, shell=True, stdout=subprocess.PIPE).stdout.decode('utf-8').split()
                return []

    def execute_dest(self, cmd, skip_logging=False, dryrun=False):
        """Execute command without output on DEST
        """
        if self.dest.location == 'ssh':
            cmd = self.ssh_string('dest') + " \"" + cmd + "\""

        # Execute the command and capture output to python list
        if not skip_logging: log.bullet4(cmd)
        if not dryrun: return subprocess.run(cmd, shell=True, stdout=subprocess.PIPE).stdout.decode('utf-8').split()

    def path(self, path=None):
        """Get the DEST server backup path (without snapshot or current, just server backup root)
        """
        result = self.dest.path + '/' + self.src.server
        if path: return result + '/' + path
        return result

    def ssh_string(self, location):
        """Get an ssh connection string for src or dest
        """
        if location == 'src':
            return ("ssh"
                + " -p " + str(self.src.ssh.port)
                + " -i " + self.src.ssh.key
                + " -o LogLevel=quiet "
                + self.src.ssh.user + "@" + self.src.ssh.host
                + " "
            )
        elif location == 'dest':
            return ("ssh"
                + " -p " + str(self.dest.ssh.port)
                + " -i " + self.dest.ssh.key
                + " -o LogLevel=quiet "
                + self.dest.ssh.user + "@" + self.dest.ssh.host
                + " "
            )



from datetime import date
from mreschke.serverbackups import Backups, log

# Optionally configure logger, but this app may have its own log settings
log.init({
    'console': {
        'level': 'INFO',  # DEBUG, INFO, WARNING, ERROR, CRITICAL
    },
    'file': {
        'enabled': True,
        'file': '/tmp/backups-{}.log'.format(date.today().strftime('%Y-%m-%d')),
    }
})

# Set per server defaults.  Each server can override parts of this default.
defaults = {
    'enabled': True,
    'cluster': None,

    # Keep all daily up to X days back, last snapshot in a week X weeks back, last snapshot in a month X months back...
    'prune': {
        'keepDaily': 30,
        'keepWeekly': 24,
        'keepMonthly': 60,
        'keepYearly': 10,
    },

    # Rsync options
    'rsync': {
        'verbose': True,
    },

    # Source server connection details
    'source': {
        'location': 'local',  # local, ssh,
        'ssh': {
            # Only used with location=ssh
            'host': 'server.example.com',  # or IP address,
            'port': 22,
            'user': 'root',
            'key': '~/.ssh/id_rsa',
        },
    },

    # Destination server connection details
    'destination': {
        'location': 'local',  # local, ssh,
        'path': '/mnt/backups',
        'ssh': {
            # Only used with location=ssh
            'host': 'server.example.com',  # or IP address,
            'port': 22,
            'user': 'root',
            'key': '~/.ssh/id_rsa',
        },
    },

    # Backup items
    'backup': {
        # Pre scripts run on destination before this servers backups, a good place to prep files for backup
        'preScripts': {
            'gitlab': {'script': 'sudo gitlab-backup create', 'output': 'gitlab-backup.txt', 'enabled': False},
        },

        # Files to backup on the destination.  Use extra in your actual server definition for extra files.
        # Arrays are merged with defaults, so you could append to common and exclude as well
        'files': {
            'common': [
                '/etc/',
                '/usr/local/bin/',
                '/root/',
                '/home/',
                '/var/log/',
                '/var/spool/cron/',
            ],
            'extra': [],
            'exclude': [
                'lost+found',
                '.Trash-1000',
            ],
        },

        # MySQL Databases and Tables
        # Host and port are relative inside the server itself via ssh, not remote
        'mysql': {
            'enabled': False,
            'mysqlCmd': 'mysql',
            #'mysqlCmd': 'docker exec -i mysql mysql',
            'dumpCmd': 'mysqldump',
            'dumpFlags': '--quick --single-transaction --flush-logs',
            #'dumpCmd': 'docker exec -i mysql mysqldump',
            'host': '127.0.0.1',
            'port': 3306,
            'user': 'root',
            'password': 'unknown',
            # All databases minus exclusions
            'dbs': '*',
            # A list of databases (all tables)
            # 'dbs': ['db1', 'db2'],
            # A list of database with a selection of tables
            # 'dbs': [
            #     {'name': 'db1', 'tables': ['table1', 'table2']},
            #     {'name': 'db2', 'tables': '*'},
            # ],
            'excludeDbs': [
                'information_schema',
                'performance_schema',
                'lost+found',
                '#mysql50#lost+found',
            ],
        },

        # Post scripts run on destination after this servers backups, a good place to cleanup files
        'postScripts': {
            'dpkg': {'script': 'dpkg -l | gzip', 'output': 'dpkg.gz', 'enabled': False},
        },
    },
}

# Define each server to backup (each will be merged with defaults config)
servers = {
    'myserver.example.com': {
        'enabled': True,
        'cluster': 'localhost',
        # Example of local -> local backup
        'source': {'location': 'local'},
        'destination': {'location': 'local'},
        'backup': {
            'preScripts': {
                #'gitlab': {'enabled': True}
                # Example of capturing script output to snapshot
                'test2': {'script': 'ls -Hhal /etc/', 'output': 'etc.txt', 'enabled': True},
            },
            'mysql': {
                'enabled': True,
                'mysqlCmd': 'mysql',
                #'mysqlCmd': 'docker exec -i mysql mysql',
                'dumpCmd': 'mysqldump',
                #'dumpCmd': 'docker exec -i mysql mysqldump',
                'host': '127.0.0.1',
                'port': 3306,
                'user': 'root',
                'password': 'password',
                'dbs': '*',
            },
            'postScripts': {
                # Export dpkg package list of Linux Debian
                'dpkg': {'enabled': True}
                # Example of backing up postgres database
                #'fusionauth': {'script': 'cd /tmp/ && sudo -u postgres pg_dump fusionauth | gzip', 'output': 'fusionauth.sql.gz', 'enabled': True},
            }
        },
    },
}

# Initialize Backup Class
# If you are defining servers in a dictionary above, without a separate defaults
#backups = Backups(servers=servers)

# If you are defining both servers and defaults above
backups = Backups(servers=servers, defaults=defaults)

# If you are using a config.d directory (/etc/serverbackups by default)
#backups = Backups()
#backups = Backups(config_path='/etc/alternative/dir')

# Run backups
# Can just run them all
backups.run()

# Or can flip through each and add your own logic per server if needed
#for server in backups.servers:
    #backups.run(server)




import logging
import logging.config
import re
import sys
from logging import Formatter

from colored import attr, bg, fg


# Usage: from mreschke.serverbackups import log
# log.init(config_dict)
# log.debug(), log.info()...standard python logging methods
def init(config={}):
    # Default Config
    # Levels = NOTSET, DEBUG, INFO, WARNING, ERROR, CRITICAL
    default = {
        'console': {
            'enabled': True,
            'level': 'INFO',
            'colors': True,
            'format': '%(message)s'
        },
        'file': {
            'enabled': False,
            'level': 'INFO',
            'file': '/tmp/example.log',
            'format': '%(asctime)s.%(msecs)03d | %(levelname)-8s | %(name)s | %(message)s'
        }
    }

    # Merge default and user defined config
    config = {**default, **config}
    if 'console' in config.keys(): config['console'] = {**default['console'], **config['console']}
    if 'file' in config.keys(): config['file'] = {**default['file'], **config['file']}

    # New Logger
    logger = logging.getLogger()
    logger.setLevel(logging.DEBUG)  # from .env

    # New Console Handler
    if config['console']['enabled']:
        ch = logging.StreamHandler(stream=sys.stdout)
        ch.setLevel(config['console']['level'])
        if config['console']['colors']:
            ch.setFormatter(ColoredFormatter(config['console']['format']))
        else:
            ch.setFormatter(logging.Formatter(
                fmt=config['console']['format'],
                datefmt='%Y-%m-%d %H:%M:%S'
            ))
        logger.addHandler(ch)

    # New File Handler
    if config['file']['enabled']:
        fh = logging.FileHandler(filename=config['file']['file'], mode='a')
        fh.setLevel(config['file']['level'])
        fh.setFormatter(logging.Formatter(
            fmt=config['file']['format'],
            datefmt='%Y-%m-%d %H:%M:%S'
        ))
        logger.addHandler(fh)


def debug(message):
    logging.debug(message)


def info(message):
    logging.info(message)


def warning(message):
    logging.warning(message)


def error(message):
    logging.error(message)


def critical(message):
    logging.critical(message)


def exception(message):
    logging.exception(message)


def header(message):
    logging.info(":: " + message + " ::")


def header2(message):
    logging.info("## " + message + " ##")


def header3(message):
    logging.info("=== " + message + " ===")


def header4(message):
    logging.info("---- " + message + " ----")


def bullet(message):
    logging.info("* " + message)


def bullet2(message):
    logging.info("- " + message)


def bullet3(message):
    logging.info("+ " + message)


def bullet4(message):
    logging.info("> " + message)


def notice(message):
    logging.info("NOTICE: " + message)


def blank():
    logging.info('')


def separator():
    logging.info('=' * 80)


def line():
    logging.info('-' * 80)


class ColoredFormatter(Formatter):

    def __init__(self, patern):
        Formatter.__init__(self, patern)

    def format(self, record):
        # Remember this is console output only, not file or other handlers
        # See color chart https://pypi.org/project/colored/
        level = record.levelname
        message = logging.Formatter.format(self, record)

        # Format header
        if (level == 'INFO' and re.match("^:: ", message)):
            message = re.sub("^:: ", "", message)
            message = re.sub(" ::$", "", message)
            message = ('{0}{1}{2}{3}').format(fg('dark_orange'), attr('bold'), ':: ', attr(0)) \
                + ('{0}{1}{2}{3}').format(fg('green'), attr('bold'), message, attr(0)) \
                + ('{0}{1}{2}{3}').format(fg('dark_orange'), attr('bold'), ' ::', attr(0))

        # Format header2
        if (level == 'INFO' and re.match("^## ", message)):
            message = re.sub("^## ", "", message)
            message = re.sub(" ##$", "", message)
            message = ('{0}{1}{2}{3}').format(fg('dark_orange'), attr('bold'), '## ', attr(0)) \
                + ('{0}{1}{2}{3}').format(fg('green'), attr('bold'), message, attr(0)) \
                + ('{0}{1}{2}{3}').format(fg('dark_orange'), attr('bold'), ' ##', attr(0))

        # Format header3
        if (level == 'INFO' and re.match("^=== ", message)):
            message = re.sub("^=== ", "", message)
            message = re.sub(" ===$", "", message)
            message = ('{0}{1}{2}{3}').format(fg('dark_orange'), attr('bold'), '=== ', attr(0)) \
                + ('{0}{1}{2}').format(fg('green'), message, attr(0)) \
                + ('{0}{1}{2}{3}').format(fg('dark_orange'), attr('bold'), ' ===', attr(0))

        # Format header4
        if (level == 'INFO' and re.match("^---- ", message)):
            message = re.sub("^---- ", "", message)
            message = re.sub(" ----$", "", message)
            message = ('{0}{1}{2}{3}').format(fg('dark_orange'), attr('bold'), '---- ', attr(0)) \
                + ('{0}{1}{2}').format(fg('green'), message, attr(0)) \
                + ('{0}{1}{2}{3}').format(fg('dark_orange'), attr('bold'), ' ----', attr(0))

        # Format bullet * item
        elif (level == 'INFO' and re.match("^\* ", message)):
            message = re.sub("^\* ", "", message)
            message = ('{0}{1}{2}').format(fg('green'), '   * ', attr(0)) \
                + ('{0}{1}{2}{3}').format(fg('white'), attr('bold'), message, attr(0))

        # Format bullet2 - item
        elif (level == 'INFO' and re.match("^- ", message)):
            message = re.sub("^- ", "", message)
            message = ('{0}{1}{2}').format(fg('cyan'), '   - ', attr(0)) \
                + ('{0}{1}{2}{3}').format(fg('white'), attr('bold'), message, attr(0))

        # Format bullet3 + item
        elif (level == 'INFO' and re.match("^\+ ", message)):
            message = re.sub("^\+ ", "", message)
            message = ('{0}{1}{2}').format(fg('red'), '   + ', attr(0)) \
                + ('{0}{1}{2}{3}').format(fg('white'), attr('bold'), message, attr(0))

        # Format bullet4 > item
        elif (level == 'INFO' and re.match("^> ", message)):
            message = re.sub("^> ", "", message)
            message = ('{0}{1}{2}').format(fg('magenta'), '   > ', attr(0)) \
                + ('{0}{1}{2}{3}').format(fg('white'), attr('bold'), message, attr(0))

        # Format notice
        elif (level == 'INFO' and re.match("^NOTICE: ", message)):
            message = re.sub("^NOTICE: ", "", message)
            message = ('{0}{1}{2}{3}').format(fg('yellow'), attr('bold'), 'NOTICE: ', attr(0)) \
                + ('{0}{1}{2}{3}').format(fg('white'), attr('bold'), message, attr(0))

        # Format separator
        elif (level == 'INFO' and re.match("^====", message)):
            message = ('{0}{1}{2}{3}').format(fg('orange_4a'), attr('bold'), message, attr(0))

        # Format line
        elif (level == 'INFO' and re.match("^----", message)):
            message = ('{0}{1}{2}{3}').format(fg('orange_4a'), attr('bold'), message, attr(0))

        elif (level == 'DEBUG'):
            message = ('{0}{1}{2}').format(fg(241), message, attr(0))
        elif (level == 'INFO'):
            message = message
        elif (level == 'WARNING'):
            message = ('{0}{1}{2}').format(fg('orange_red_1'), message, attr(0))
        elif (level == 'ERROR'):
            message = ('{0}{1}{2}').format(fg('red'), message, attr(0))
        elif (level == 'CRITICAL'):
            message = ('{0}{1}{2}{3}').format(fg('white'), bg('red'), message, attr(0))

        return message



from setuptools import setup

with open("README.md", "r") as fh:
    long_description = fh.read()

setup(
    name="mreschke-serverbackups",
    version="1.0.3",
    description="Config driven server backups that feel like a simple backup.py script",
    long_description=long_description,
    long_description_content_type="text/markdown",

    author="Matthew Reschke",
    author_email="mail@mreschke.com",
    license="MIT",
    url="http://github.com/mreschke/server-backups",

    # I am using Native namespaces see https://packaging.python.org/guides/packaging-namespace-packages/#native-namespace-packages
    packages=["mreschke.serverbackups"],
    package_data={
        'mreschke.serverbackups': ['templates/*'],
    },
    zip_safe=False,
    python_requires=">=3.5",
    classifiers=[
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "License :: OSI Approved :: MIT License",
        "Operating System :: Unix",
        "Environment :: Console",
    ],
    install_requires=[
        "click==7.1.*",
        "colored==1.4.*",
        "prettyprinter==0.18.*",
        "PyYAML==5.*",
    ],
    entry_points={
        "console_scripts": [
            "serverbackups=mreschke.serverbackups.init:cli"
        ],
    },
)



import os
import click
from shutil import copyfile

CONTEXT_SETTINGS = dict(help_option_names=['-h', '--help'])


@click.group(context_settings=CONTEXT_SETTINGS)
def cli():
    """
    \b
    mreschke.serverbackups
    Copyright (c) 2018 Matthew Reschke
    License http://mreschke.com/license/mit
    """
    pass


@cli.command('init-cli-script')
@click.option('-f', '--filename', default='backups.py', show_default=True)
def init_cli_script(filename):
    """Initialize a new backups.py in the current folder
    """
    template = os.path.dirname(os.path.realpath(__file__)) + '/templates/' + 'backups_cli.py'
    path = os.getcwd()
    file = path + '/' + filename

    if os.path.isfile(file): exit("File " + filename + " already exists")
    if click.confirm("You are about to create a new " + file + ", continue?"):
        click.secho("Creating new backup file " + file, fg='green')
        copyfile(template, file)
        click.secho("Done")
    else:
        exit("Cancelled, bye.")


@cli.command('init-directory')
@click.option('-p', '--path', default='/etc/serverbackups', show_default=True)
def init_directory(path):
    """Initialize a backup and config.d directory
    """
    defaults = os.path.dirname(os.path.realpath(__file__)) + '/templates/' + 'defaults.yml'
    server = os.path.dirname(os.path.realpath(__file__)) + '/templates/' + 'server_configd.yml'

    if os.path.exists(path):
        exit("Path already exists")

    if click.confirm("You are about to create a new " + path + ", continue?"):
        click.secho("Creating new backup directory " + path, fg='green')
        os.makedirs(path)
        os.mkdir(path + '/config.d')
        copyfile(defaults, path + '/defaults.yml')
        copyfile(server, path + '/config.d/myserver.example.com.yml')
        click.secho("Done")
    else:
        exit("Cancelled, bye.")


@cli.command('example-api-module')
def example_api_module():
    """Example of using the Backup class in your own code
    """
    template = os.path.dirname(os.path.realpath(__file__)) + '/templates/' + 'backups_api.py'
    with open(template, 'r') as f:
        print(f.read())


@cli.command('example-defaults')
def example_defaults():
    """Example of a default.yml config
    """
    template = os.path.dirname(os.path.realpath(__file__)) + '/templates/' + 'defaults.yml'
    with open(template, 'r') as f:
        print(f.read())


@cli.command('show-builtin-defaults')
def show_defaults():
    """Builtin sensible defaults applied when you dont
    """
    template = os.path.dirname(os.path.realpath(__file__)) + '/templates/' + 'defaults_blank.yml'
    with open(template, 'r') as f:
        print(f.read())


# Initiate cli
if __name__ == '__main__':
    cli()



import sys

import click

from . import log
from .utils import dd, dump

# Exit if not running as root - NO, can be use backups too
# if os.geteuid() != 0: exit('This script must be run as root')

CONTEXT_SETTINGS = dict(help_option_names=['-h', '--help'])


@click.group(context_settings=CONTEXT_SETTINGS)
def handle():
    """
    \b
    mreschke.serverbackups
    Copyright (c) 2018 Matthew Reschke
    License http://mreschke.com/license/mit
    """
    pass


@handle.command()
# @click.option('-a', '--all', default='backups.py', show_default=True)
@click.option('--all', is_flag=True, help="Run all backups (both config.yml and custom code)")
@click.option('--custom', is_flag=True, help="Run only custom backup code")
@click.option('--servers', is_flag=True, help="Run entire config.yml only, no custom code")
@click.option('--server', help="Run just one server from config.yml")
@click.option('--cluster', help="Run all servers in a cluster from config.yml")
def run(all, custom, servers, server, cluster):
    """Run backups"""


def start(backups):
    # Only run with proper cli flags
    if 'run' not in sys.argv and '--help' not in sys.argv:
        log.error('Missing run option, see --help')
        exit(1)

    if not any(x in ['--all', '--servers', '--server', '--cluster'] for x in sys.argv):
        if 'run' in sys.argv and '--custom' not in sys.argv and '--help' not in sys.argv and '-h' not in sys.argv:
            log.error('Missing run option, see --help')
            exit(1)
        return

    log.blank()
    log.separator()
    log.header3("Running server backups")
    log.separator()

    if backups.servers:
        # Passing in a dictionary of servers
        log.bullet("Backups defined as inline dictionary")
    else:
        # Using config directory
        log.bullet("Using config dir {}".format(backups.config_path))

    # Run a single server
    if '--server' in sys.argv:
        server = sys.argv[sys.argv.index("--server") + 1]
        log.bullet("Running single server '{}'".format(server))
        backups.run(server=server)

    # Run all servers in a single cluster
    elif '--cluster' in sys.argv:
        cluster = sys.argv[sys.argv.index("--cluster") + 1]
        log.bullet("Running servers in cluster '{}'".format(cluster))
        backups.run(cluster=cluster)

    else:
        log.bullet("Running all servers defined in config")
        backups.run()


# def backupserversOLD(config_path='/etc/mreschke/serverbackups', servers=None, defaults=None):
#     """Run the main BackupServer class from the server config.yml and config.d files
#     """
#     # Only run with proper cli flags
#     if not any(x in ['--all', '--servers', '--server', '--cluster'] for x in sys.argv): return

#     log.blank()
#     log.separator()
#     log.header3("Running server backups")
#     log.separator()

#     if servers:
#         # Passing in a dictionary of servers
#         log.bullet("Backups defined as inline dictionary")
#         if not defaults:
#             log.bullet("No DEFAULTS dictionary passed inline, using builtin DEFAULTS")
#             defaults_file = os.path.dirname(os.path.realpath(__file__)) + '/templates/' + 'defaults.yml'
#             with io.open(defaults_file, 'r') as stream:
#                 defaults = yaml.load(stream, Loader=yaml.SafeLoader)
#         else:
#             log.bullet("DEFAULTS defined as inline dictionary string")
#             defaults = yaml.load(defaults)
#     else:
#         # Using config directory
#         log.bullet("Using config dir {}".format(config_path))
#         configd_path = config_path + '/config.d'
#         defaults_file = config_path + '/defaults.yml'

#         # Read defaults.yml file
#         with io.open(defaults_file, 'r') as stream:
#             defaults = yaml.load(stream)

#         # Read each server config in config.d folder
#         servers = {}
#         for file in os.listdir(configd_path):
#             with io.open(configd_path + "/" + file, 'r') as stream:
#                 servers = {**servers, **yaml.load(stream)}

#     # Run a single server
#     if '--server' in sys.argv:
#         server = sys.argv[sys.argv.index("--server") + 1]
#         log.bullet("Running single server '{}'".format(server))
#         BackupServer(log, server, servers[server], defaults).run()

#     # Run all servers in a single cluster
#     elif '--cluster' in sys.argv:
#         cluster = sys.argv[sys.argv.index("--cluster") + 1]
#         log.bullet("Running servers in cluster '" + cluster + "'")
#         for server, options in servers.items():
#             if options['cluster'] == cluster:
#                 BackupServer(log, server, options, defaults).run()

#     else:
#         log.bullet("Running all servers defined in config")
#         for server, options in servers.items():
#             BackupServer(log, server, options, defaults).run()

def allowcustom():
    """Helper for backups.py to determine if custom backup code should be run
    """
    return any(x in ['--custom', '--all'] for x in sys.argv)


# Explicit exports
__all__ = ['handle', 'allowcustom', 'start', 'log', 'dump', 'dd']



from prettyprinter import cpprint


def dump(*args):
    """Dump variables using prettyprinter"""
    for arg in args:
        cpprint(arg)


def dd(*args):
    """Dump variables using prettyprinter and exit()"""
    for arg in args:
        cpprint(arg)
    exit()
